# （1）拡張モジュールのインポート・文字のシンボル化
import numpy as np                  # 配列を扱う数値計算NumPyモジュール エイリアスnp
import numpy as np                  # 配列を扱う数値計算NumPyモジュール エイリアスnp
import matplotlib.pyplot as plt     # グラフ描画Matplotlibモジュール エイリアスplt
import japanize_matplotlib          # Matplotlibの日本語化モジュール
import sympy as sym                 # 代数計算（数式処理）SymPyモジュール エイリアスsym
# 厳密解計算sym.dsoveのためシンボリック変数・関数を設定
t = sym.Symbol('t')                 # 全モデル共通シンボリック変数 時間変数t
N, k, N_0 = sym.symbols('N k N_0')  # モデル1 シンボリック変数 メイン変数N(人口), パラメーターk、初期値N_0
x, k, x_0 = sym.symbols('x k x_0')  # モデル2 シンボリック変数 メイン変数x(企業数), パラメーターk, 初期値x_0
w = sym.Symbol('w')                 # モデル3 シンボリック変数 メイン変数w(魚の重さ)
f = sym.Function('f')               # 全モデル共通シンボリック関数

# （2）モデルの選択
# モデル1:マルサス・モデル N:人口 , t:時間
# モデル2:技術革新の普及モデル x:技術革新を採り入れた企業の数 , t:時間
# モデル3:魚の成長モデル x:魚の体重 , t:時間
while(1):
  Model = input('モデル1：マルサス・モデル\r\n'
                'モデル2：技術革新の普及モデル\r\n'
                'モデル3：魚の成長モデル\r\n'
                '1、2、3のどれかを入力>>> ')
  if Model.isdecimal():
    Model = int(Model)
    if 1 <= Model <= 3:
      break

# （3）モデルの定義
# 微分方程式
# sym.dsolveによる厳密解の計算
# パラメーター
# 変数tの範囲
# 数値解法のため関数dXdT(X, T)の定義
# グラフタイトルと軸ラベル単位設定

# モデル1:マルサス・モデル N:人口 , t:時間
if Model == 1:
  # 微分方程式 dN/dt=kN
  dNdt = k*N    # シンボリック変数:k, N
  # 厳密解計算
  # 準備：Nをf(t)に置き換える dNdt = k*N -> f'(t) = k*f(t)
  deq = sym.Eq(sym.Derivative(f(t), t), dNdt.subs(N, f(t)) )    # シンボリック関数:f,シンボリック変数:t
  gen = sym.dsolve(deq, ics={f(0):N_0})     # 厳密解gen N_0:シンボリック変数
  genrr = gen.rhs                           # genrr = 厳密解の【右辺】（kとN_0はそのまま）
  # パラメーターkの値
  kk = 0.306
  # 初期条件 N(0) = N_0(シンボリック変数) = NN_0 (変数) = 5300000 (1800年の米国人口)
  NN_0 = 5.3e+6
  # 変数tの範囲 0≦t=t_[j]≦t_max
  t_max = 10
  # 厳密解数値計算用関数genr
  genr = genrr.subs([(k,kk),(N_0, NN_0)])      # 厳密解の右辺genrrのkに数値(kk)、初期条件 N(0)にN_0を代入
  # 数値解法用関数dXdT(X, T)の定義、数値解法用初期値x0
  dXdT = lambda X,T : dNdt.subs([(N, X),(t, T),(k, kk)])
  x0 = NN_0
  # グラフの大きさを指定
  plt.figure(figsize=(12.0, 7.0))
  # 1800年から1900年までの米国人口統計 データプロット
  data = [[0, 5.3e+6],[1, 7.2e+6],[2, 9.6e+6],[3, 12.9e+6],[4, 17.1e+6],[5, 23.2e+6],\
         [6, 31.4e+6],[7, 38.6e+6],[8, 50.2e+6],[9, 62.9e+6],[10, 76.0e+6]]
  tt, NN = zip(*data)
  plt.scatter(tt, NN, color = "red" , label = "米国人口統計")
  # グラフ設定
  plt.title("モデル1 マルサス・モデル 微分方程式 dN/dt=kN , k=" + str(kk) + " , 初期条件 N(0)=N_0=" + \
            '{:.0f}'.format(x0) + "の解曲線（厳密解と数値解）")
  plt.xlabel('西暦(時間変数t)')                     # x軸ラベル
  plt.xticks(np.arange(0, t_max+0.1, t_max/10))     # x軸目盛り 10等分
  # t=0を1800年、t=10を1900年に対応させる
  plt.gca().set_xticklabels(['{:.0f}年'.format(1800+10*xx)+'({:.0f})'.format(xx) for xx in plt.gca().get_xticks()])
  plt.ylabel('人口N（人）')                             # y軸ラベル
  plt.yticks(np.arange(0,1100000000+0.1 , 10000000))    # y軸目盛り 0から1100万までを100万刻み
  plt.gca().set_yticklabels(['{:.0f}万'.format(y/10000) for y in plt.gca().get_yticks()])   # y軸目盛り表示調整

# モデル2 技術革新の普及モデル x:技術革新を採り入れた企業の数 , t:時間 , N:コミュニティ内の企業数
elif Model == 2:
  # 微分方程式 dx/dt=kx(N-x)
  N = 100             # コミュニティ内の企業数Nは100に固定
  dxdt = k*x*(N-x)    # シンボリック変数:x, k
  # 厳密解計算
  # 準備：xをf(t)に置き換える dxdt = k*x*(N-x) -> f'(t) = k*f(t)*(N-f(t))
  deq = sym.Eq(sym.Derivative(f(t), t), dxdt.subs(x, f(t)))  # シンボリック関数:f,シンボリック変数:t
  gen = sym.dsolve(deq, f(t), ics={f(0):x_0})      # 厳密解gen x_0:シンボリック変数
  genrr = gen.rhs                                  # genrr = 厳密解genの右辺gen.rhs（k,x_0はそのまま）
  # 初期条件 x(0)=1
  x0 = 1
  # パラメーターkの値
  kk = 0.03     # k:正の定数
  # 変数tの範囲 0≦t=t_[j]≦t_max
  t_max = 5
  # 厳密解数値計算用関数genr
  genr = genrr.subs([(k,kk),(x_0, x0)])    # 厳密解の右辺genrrのkに数値(kk)、初期条件 x_0に数値x0を代入
  # 数値解法用関数dXdT(X, T)の定義、数値解法用初期値x0
  dXdT = lambda X,T : dxdt.subs([(x, X),(t, T),(k, kk)])
  x0 = x0
  # グラフの大きさを指定
  plt.figure(figsize=(12.0, 7.0))
  # グラフ設定
  plt.title("モデル2 技術革新の普及モデル 微分方程式 dx/dt=k*x*(100-x), k=" + str(kk)+ ", 初期条件 x(0)=" + \
            '{:.0f}'.format(x0) + "の解曲線（厳密解と数値解）")
  plt.xlabel('t(年)')               # x軸ラベル
  plt.ylabel('x(企業数)')           # y軸ラベル
  plt.yticks(np.arange(0,N+0.1,N/10))    # y軸目盛り 0からNN(100)までを10分の1刻み

# モデル3 魚の成長モデル w:魚の体重(g) , t:時間
elif Model == 3:
  # パラメーター
  a = 5      # 体重増加要因係数(a>0)
  b = 1      # 体重減少要因係数(b>0)
  # 微分方程式  dw/dt = aw^(2/3) - bw
  dwdt = a*w**(2/3) - b*w    # シンボリック変数:w
  # 初期条件  w(0)=0
  w0 = 0
  # 厳密解計算
  # 準備：wをf(t)に置き換える dwdt = a*w**(2/3) - b*w -> f'(t) = a*f(t)**(2/3)-b*f(t)
  deq = sym.Eq(sym.Derivative(f(t), t), dwdt.subs(w, f(t)))     # シンボリック関数:f,シンボリック変数:t
  gen = sym.dsolve(deq, ics={f(0):w0}, hint="Bernoulli")   # 厳密解gen Bernoulli型の解法により解く
  genr = gen.rhs                                           # genr = 厳密解genの右辺gen.rhs
  # 変数tの範囲 0≦t=t_[j]≦t_max  (0≦t≦20)
  t_max = 20
  # 数値解法用関数dXdT(X, T)の定義、数値解法用初期値x0
  dXdT = lambda X,T : dwdt.subs([(w, X),(t, T)])
  x0 = 0.001    # x0=0 とすると数値解法の結果はw(t)=0となってしまうため0に近い値に設定
  # グラフの大きさを指定
  plt.figure(figsize=(12.0, 7.0))
  # グラフ設定
  plt.title("モデル3 魚の成長モデル 微分方程式 dw/dt=" + str(dwdt) + " , 初期条件 w(0)=" + \
            '{:.0f}'.format(x0) + "の解曲線（厳密解と数値解）")
  plt.xlabel('t(カ月)')            # x軸ラベル
  plt.ylabel('体重 w(g)')          # y軸ラベル

# （4）変数tの定義域設定・リストt_[j]生成
T = t_max                     # 変数tの範囲 0≦t=t_[j]≦T  (0≦t≦20)
n = 100                       # 0≦t=t_[j]≦t_maxをn等分  (100等分)
h = T / n                     # 等差数列の公差（刻み幅） (h=20/100=0.2)
t_ = np.linspace(0, T, n + 1) # numpy.linspace(初項, 末項, 項数)等差数列のリストを生成
# 0から公差hでTまでの範囲で等差数列を生成   t_[0],...,t_[n]   要素数n+1

# （5）数値解法用のxeとx4、厳密解用のxgをリストとして生成
xe = np.empty(n + 1)        # オイラー法用のリスト           xe[0],...,xe[n] 要素数n+1個
x4 = np.empty(n + 1)        # 4次ルンゲ・クッタ法の用リスト  x4[0],...,x4[n] 要素数n+1個
xg = np.empty(100 + 1)      # 厳密解用のリスト               xg[0],...,xg[100] 要素数100+1個

# 初期値x0を代入
xe[0] = x0
x4[0] = x0
xg[0] = x0

# （6）数値解法
for j in range(n):        # j=0,...,n-1 -> x@[0],...,x@[n] (@=e,4,g) 要素数n+1個
# オイラー法 Euler Methods
  xe[j+1] = xe[j] + h * dXdT( xe[j], t_[j] )                  # 末項 j=n-1 -> xe[j+1] = xe[n]
# 4次ルンゲ・クッタ法 4th-Order Runge-Kutta Methods
  k4X1 = h * dXdT( x4[j], t_[j] )
  k4X2 = h * dXdT( x4[j] + k4X1/2, t_[j] + h/2 )
  k4X3 = h * dXdT( x4[j] + k4X2/2, t_[j] + h/2 )
  k4X4 = h * dXdT( x4[j] + k4X3, t_[j] + h )
  x4[j+1] = x4[j] + 1/6 * ( k4X1 + 2*k4X2 + 2*k4X3 + k4X4 )   # 末項 j=n-1 -> x4[j+1] = x4[n]

# （7）厳密解のグラフをプロット表示させるための数値計算
tt_ = np.linspace(0, t_max, 100 + 1)
for j in range(100):
  xg[j+1] = genr.subs(t, tt_[j+1])

# （8）グラフ表示
# 点(t,x) n+1個をプロット
plt.plot(t_, xe, color = "blue", label = "オイラー法 （n=" + str(n) + "）", linewidth = 1)
plt.plot(t_, x4, color = "green", label = "4次ルンゲ・クッタ法 （n=" + str(n) + "）", linewidth = 3)
plt.plot(tt_, xg, color = "red", label = "厳密解 x(t) = " + str(genr), linewidth = 1)
# グラフの見た目設定
plt.grid(True)              # グリッド表示
plt.legend()                # 凡例表示
# 設定を反映してプロット描画
plt.show()
